<?php
namespace Zeedhi\PaymentPlatform\Implementations\SoftwareExpress;

use Zeedhi\PaymentPlatform\BankSlip\BankSlipPlatform;
use Zeedhi\PaymentPlatform\BankSlip\BankSlipTransaction;

use Zeedhi\PaymentPlatform\InstantBuy\InstantBuyPlatform;
use Zeedhi\PaymentPlatform\InstantBuy\InstantBuyTransaction;

use Zeedhi\PaymentPlatform\NonInstantBuy\NonInstantBuyPlatform;
use Zeedhi\PaymentPlatform\NonInstantBuy\NonInstantBuyTransaction;

use Zeedhi\PaymentPlatform\Consult\ConsultPlatform;
use Zeedhi\PaymentPlatform\Consult\Consult;

use Zeedhi\PaymentPlatform\Revoke\RevokePlatform;
use Zeedhi\PaymentPlatform\Revoke\Revoke;

use Zeedhi\PaymentPlatform\CreditCard;
use Zeedhi\PaymentPlatform\Customer;
use Zeedhi\PaymentPlatform\RequestURL;

use Zeedhi\PaymentPlatform\Exception;

class PaymentImpl implements InstantBuyPlatform, NonInstantBuyPlatform{
    const SANDBOX = 'SANDBOX';
    const ENDPOINT_SANDBOX = "https://esitef-homologacao.softwareexpress.com.br/e-sitef-hml/api";

    const PRODUCTION = 'PRODUCTION';
    const ENDPOINT_PRODUCTION = "https://esitef-ec.softwareexpress.com.br/e-sitef/api";

    const MAXREDIRS = 10;
    const TIMEOUT = 30;
    const POST = "POST";
    const PUT = "PUT";

    /** @var String */
    private $merchantKey;

    /** @var array */
    private $requestHeaders;

    /** @var HttpRequest */
    private $httpRequest;

    /** @var string */
    private $urlRequest;

    public function __construct($merchantKey, $merchantId, HttpRequest $httpRequest, $urlRequest) {

        $this->merchantKey = $merchantKey;
        $this->merchantId = $merchantId;
        $this->httpRequest = $httpRequest;
        $this->urlRequest = $urlRequest;
        $this->requestHeaders = array(
            'Content-Type: application/json',
            'merchant_id: '.$this->merchantId,
            'merchant_key: '.$this->merchantKey
        );

    }

    public function createToken(CreditCard $card, Customer $customer) {
        $invalidChars = array('20', '/', ':', ' ');
        $merchantUSN = str_replace($invalidChars, '', date('Y/m/d H:i:s'));
        $request = array(
            'card' => array(   
                "holder"     => $card->getHolderName(),
                'expiry_date' => $card->getExpDateMonth().''.$card->getExpDateYear(),
                'number' => $card->getNumbers()
            ),
            'authorizer_id' => $this->decodeFlag($card->getFlag()),
            'merchant_usn' => $merchantUSN,
            'customer_id' => $customer->getId()
        );
        $responseData = $this->createRequest(RequestURL::SOFTWAREEXPRESS_CARDS, self::POST, $request);
        return array(
            'InstantBuyKey' => $responseData['card']['token']
        );
    }

    public function createTransaction($orderId, $installments, $installmentType, $authorizerId, $amount){
        $invalidChars = array('20', '/', ':', ' ');
        $merchantUSN = str_replace($invalidChars, '', date('Y/m/d H:i:s'));
        $request = array(
            'merchant_usn' => $merchantUSN,
            'order_id' => $orderId,
            'installments' => $installments , 
            'installment_type' => $installmentType, 
            'authorizer_id' => $authorizerId,
            'amount' => $amount,
            'additional_data' => array(
                'postpone_confirmation' => 'true'
            )
        );
        $responseData = $this->createRequest(RequestURL::SOFTWAREEXPRESS_TRANSACTIONS, self::POST, $request);
        
        return $responseData['payment']['nit'];
    }

    public function createBuyTransaction($nit, $request){
        $responseData = $this->createRequest(RequestURL::SOFTWAREEXPRESS_PAYMENTS.'/'.$nit, self::POST, $request);

        $captureUrl = $this->getCaptureUrl($nit);
        return array(
            'status' => $responseData['payment']['status'],
            'message' => $responseData['payment']['authorizer_message'],
            'captureUrl' => $captureUrl,
            'paymentId' => null,
            'nsu' => isset($responseData['payment']['host_usn']) ? preg_replace('/[^0-9]/', '', $responseData['payment']['host_usn']) : null
        );
    }
    
    public function instantBuyTransaction(InstantBuyTransaction $instantBuyTransaction) {
        $orderId = $instantBuyTransaction->getTransactionId();
        $installments = '1'; //Pagamento à vista
        $installmentType = '4'; // Parcelamento pela loja, usar 4 quando for pagamento à vista
        $authorizerId = $this->decodeFlag($instantBuyTransaction->getFlag());
        $amount = $instantBuyTransaction->getValue();

        $nit = $this->createTransaction($orderId, $installments, $installmentType, $authorizerId, $amount);
        
        $cvv = $instantBuyTransaction->getCvv();
        
        if($cvv === null){
            $request = array(
                "card" => array(
                    "token" => $instantBuyTransaction->getCreditCardToken()
                )
            );   
        } else{
            $request = array(
                "card" => array(
                    "token" => $instantBuyTransaction->getCreditCardToken(),
                    "security_code" => $cvv
                )
            );
        }
        
        return $this->createBuyTransaction($nit, $request);
    }

    public function nonInstantBuyTransaction(NonInstantBuyTransaction $nonInstantBuyTransaction) {
        $orderId = $nonInstantBuyTransaction->getTransactionId();
        $installments = '1'; //Pagamento à vista
        $installmentType = '4'; // Parcelamento pela loja, usar 4 quando for pagamento à vista
        $authorizerId = $this->decodeFlag($nonInstantBuyTransaction->getCreditCard()->getFlag());
        $amount = $nonInstantBuyTransaction->getValue();

        $nit = $this->createTransaction($orderId, $installments, $installmentType, $authorizerId, $amount);
        
        $request = array(
            "card"   => array(
                "holder"     => $nonInstantBuyTransaction->getCreditCard()->getHolderName(),
                "number"     => $nonInstantBuyTransaction->getCreditCard()->getNumbers(),
                "expiry_date" => $nonInstantBuyTransaction->getCreditCard()->getExpDateMonth().$nonInstantBuyTransaction->getCreditCard()->getExpDateYear(),
                "security_code"   => $nonInstantBuyTransaction->getCreditCard()->getCvv()
            )
        );

        return $this->createBuyTransaction($nit, $request);
    }

    private function getCaptureUrl ($nit) {
        return RequestURL::SOFTWAREEXPRESS_PAYMENTS.'/'.$nit.'?confirm=true';
    }

    public function captureTransaction($url, $paymentId) {
        $request = array();
        $captureResponse = $this->createRequest($url, self::PUT, $request);
        return $captureResponse;
    }

    private function createRequest($url, $verb, $fields){
        $this->configCurl($url);
        //if($this->urlRequest == self::ENDPOINT_SANDBOX) {
            $this->httpRequest->setOption(CURLOPT_PROXY, "");
            $this->httpRequest->setOption(CURLOPT_PROXYUSERPWD, "");
            $this->httpRequest->setOption(CURLOPT_SSL_VERIFYPEER, false);
        //}
        $this->httpRequest->setOption(CURLOPT_CUSTOMREQUEST, $verb);
        $this->httpRequest->setOption(CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
        $this->httpRequest->setOption(CURLOPT_MAXREDIRS, self::MAXREDIRS);
	$this->httpRequest->setOption(CURLOPT_TIMEOUT, self::TIMEOUT);
	if(!empty($fields)){
		$this->httpRequest->setOption(CURLOPT_POSTFIELDS, json_encode($fields));
	}
        return $this->responseTreatment();
    }
    
    private function responseTreatment() {
        $response = json_decode($this->httpRequest->execute(), true);
        $httpStatusCode = $this->httpRequest->getInfo(CURLINFO_HTTP_CODE);
        $this->httpRequest->close();
        if($httpStatusCode == '201' || $httpStatusCode == '200') {
            return $response;
        } else {
            throw Exception::apiSoftwareExpress($httpStatusCode, $response['message']);
        }
    }

    private function configCurl($url) {
        $this->httpRequest->setUrl($this->urlRequest.$url);
        $this->httpRequest->setOption(CURLOPT_HTTPHEADER, $this->requestHeaders);
        $this->httpRequest->setOption(CURLOPT_RETURNTRANSFER, true);
    }

    public static function factory($merchantKey, $merchantId, $flag) {
        if($flag === self::SANDBOX){
            $urlRequest = self::ENDPOINT_SANDBOX;
        }else if($flag === self::PRODUCTION){
            $urlRequest = self::ENDPOINT_PRODUCTION;
        }
        return new self($merchantKey, $merchantId, new HttpRequest, $urlRequest);
    }

    private function decodeFlag($flag) {
        switch($flag) {
            case CreditCard::FLAG_VISA:
                return 1;
            case CreditCard::FLAG_MASTERCARD:
                return 2;
            case CreditCard::FLAG_AMEX:
                return 3;
            case CreditCard::FLAG_HIPERCARD:
                return 5;
            case CreditCard::FLAG_AURA:
                return 6;
            case CreditCard::FLAG_DINERS:
                return 33;
            case CreditCard::FLAG_ELO:
                return 41;
            case CreditCard::FLAG_ALELO:
                return 224;
            case CreditCard::FLAG_SODEXO:
                return 281;
            default:
                throw Exception::unsupportedFlag();
                break;
        }
    }
}
